static void 
#ifdef _CLOVER
assign_inv_mult_eoprec_wd(
                    const float *kappa2,
                      cuhqfield *yde,
               const  cuhqfield *ye,
               const cuhgvfield *ue,
               const cuhgvfield *uo,
               const int *ieo,
               const cuhclvfield *fclinve,
               const cuhclvfield *fclinvo){
#else
assign_inv_mult_eoprec_wd(
                    const float *kappa2,
                      cuhqfield *yde,
               const  cuhqfield *ye,
               const cuhgvfield *ue,
               const cuhgvfield *uo,
               const int *ieo) {
#endif
///////////////////////////////////////////////////////////////////////////////
// 
// Multiply inverse even/odd-site preconditioned Dirac operator : 
//       v
// yde = Dee \ ye,
//       v
// where Dee = 1 - kappa^2 Meo Moe
//
// mkappa2 : hopping parameter (= -kappa^2)
//      ye : even/odd site fermion (input)
//     yde : even/odd site fermion (output)
//      ue : even/odd site gauge link
//      uo : odd/even site gauge link
//     ieo : 0 or 1 (0:even,1:odd)
// fclinve : even/odd site inverse clover term
// fclinvo : odd/even site inverse clover term
//
///////////////////////////////////////////////////////////////////////////////
  cuhqfield *Myo;
  int *ioe;

  CUDA_SAFE_CALL(cudaMalloc((void**)&Myo,sizeof(cuhqfield)));
  CUDA_SAFE_CALL(cudaMalloc((void**)&ioe,sizeof(float)));
  cureverse_eo(ieo,ioe); // ioe = (ieo+1) % 2

  assign(yde,ye);

  for (int iter=0; iter < _PREC_ITER_; iter++) {

#ifdef _CLOVER
    // Myo = F^-1_oo Moe yde
    // yde = F^-1_ee Meo Myo = F^-1_ee Meo F^-1_oo Moe yde
    assign_mult_hop(Myo,yde,uo,ue,ioe,fclinvo);
    assign_mult_hop(yde,Myo,ue,uo,ieo,fclinve);
#else
    // Myo = Moe yde
    // yde = Meo Myo = Meo Moe yde
    assign_mult_hop(Myo,yde,uo,ue,ioe);
    assign_mult_hop(yde,Myo,ue,uo,ieo);
#endif

    //
    // yde = + kappa^2 yde + ye
    //
    accum_mult_add(yde,kappa2,ye);

  }

  CUDA_SAFE_CALL(cudaFree(Myo));
  CUDA_SAFE_CALL(cudaFree(ioe));
}
