#include <config.h>
module hmc_poly_quark_wilson_class
!
!= Define HMC quark Polynomial HMC algorithm (even/odd-site preconditioned)
!
!== Version
!
! $Id: hmc_poly_quark_wilson_class.F90,v 1.26 2011/06/21 03:29:27 ishikawa Exp $
!
  use counter_class
  use hmc_logfile_class
  use error_class
  use solver_parameter_class
  use lattice_class
  use metropolis_test_class
  use hmc_status_class
  use field_gauge_class
  use field_fermion_class
#ifdef _CLOVER
  use quark_clover_class
#else
  use quark_wilson_class
#endif
  implicit none
  private
  public :: hmc_poly_quark_wilson
  public :: force
  public :: hamil
  public :: metropolis
  public :: reweighting
  public :: new
  public :: delete
  public :: read
  public :: save_config
  public :: print
  public :: print_statistics
  public :: PHMC_ALGORITHM
  public :: set_id
  public :: has_force_md_depth
  public :: has_global_metropolis_test
  public :: has_reweighting

  integer, parameter :: PHMC_ALGORITHM=0
  character(len=*), parameter :: SOLVER_LOG_FNAME_DEFAULT="wq_phmc_solver_log."
  character(len=*), parameter ::    GMP_LOG_FNAME_DEFAULT="wq_phmc_gmp_log."

  type hmc_poly_quark_wilson
  !
  ! HMC quark even/odd sitem preconditioned Polynomical HMC algorithm action
  !
  !

#ifdef _CLOVER
    type(quark_clover),     public :: phys_param
    type(tfield_gluon_wg),  public, pointer :: XX => NULL()
#else
    type(quark_wilson),     public :: phys_param
#endif

    type(field_quark_wg),   private, pointer :: y  => NULL()
    type(solver_parameter), private :: update_q_inv
    type(solver_parameter), private ::      gmp_inv

    type(hmc_logfile),      private :: solver_log
    character(CHARLEN),     private :: solver_log_fname = SOLVER_LOG_FNAME_DEFAULT

    type(metropolis_status),private :: gmp_status
    character(CHARLEN),     private ::    gmp_log_fname =    GMP_LOG_FNAME_DEFAULT

    character(CHARLEN),     private ::    gmp_test_name = "Global Metropolis"
    character(CHARLEN),     private ::  gmp_weight_name = "dS"

    real(DP),               private, allocatable :: pcoef(:) ! Nf=2       polynomial coefficients
    complex(DP),            private, allocatable :: zcoef(:) ! Nf=1 or 2, polynomial coefficients

    integer, private :: myid = 0       ! quark id
    integer, private :: nflavor = 0    ! number of flavor
    integer, private :: npoly = 0      ! polynomial order
    integer, private :: ptype = 0      ! polynomial approximation type index
    integer, private :: nsqrt = 0      ! Taylor expansion order for sqrt of correction matrix for Nf=1
    integer, private :: switch_metropolis_test = 0  ! global HMC Metropolis test switch
    integer, private :: depth_hopping  = 1
    integer, private :: depth_clvtrlog = 1
    logical, private :: has_global_metropolis_test = .true.
    logical, private :: has_reweighting            = .false.
  end type

  interface new
    module procedure new_wilson_quark_phmc
  end interface

  interface delete
    module procedure delete_wilson_quark_phmc
  end interface

  interface read
    module procedure read_wilson_quark_phmc
  end interface

  interface save_config
    module procedure save_config_hmc_poly_quark_wilson
  end interface

  interface print
    module procedure print_wilson_quark_phmc
  end interface

  interface print_statistics
    module procedure print_stat_wilson_quark_phmc
  end interface

  interface force
    module procedure force_wilson_quark_phmc
  end interface

  interface hamil
    module procedure hamil_wilson_quark_phmc
  end interface

  interface metropolis
    module procedure metropolis_wilson_quark_phmc
  end interface

  interface reweighting
    module procedure reweighting_wilson_quark_phmc
  end interface

  interface set_id
    module procedure set_id_hpqw
  end interface

  interface has_force_md_depth
    module procedure has_force_wilson_quark_phmc
  end interface

  interface has_global_metropolis_test
    module procedure has_gmp_wilson_quark_phmc
  end interface

  interface has_reweighting
    module procedure has_reweighting_wilson_quark_phmc
  end interface

contains 

subroutine set_id_hpqw(this,id)
!
! set quark # id
!
  implicit none
  type(hmc_poly_quark_wilson), intent(inout) :: this
  integer, intent(in) :: id
  this%myid = id
  call set_id(this%phys_param,id)
  return
end subroutine

function has_force_wilson_quark_phmc(this,md_depth) result(has_force)
!
! inquire whether this has force at the given md depth.
!
  implicit none
  type(hmc_poly_quark_wilson), intent(in) :: this
  integer, intent(in) :: md_depth
  logical :: has_force
  if ((md_depth == this%depth_hopping) .or. &
 &    (md_depth == this%depth_clvtrlog)) then
    has_force = .TRUE.
  else
    has_force = .FALSE.
  endif
  return
end function

function has_gmp_wilson_quark_phmc(this) result(has_gmp)
!
! inquire whether this algorithm has GMP
!
  implicit none
  type(hmc_poly_quark_wilson), intent(in) :: this
  logical :: has_gmp
  has_gmp = this%has_global_metropolis_test
  return
end function

function has_reweighting_wilson_quark_phmc(this) result(has_rew)
!
! inquire whether this algorithm has reweighting
!
  implicit none
  type(hmc_poly_quark_wilson), intent(in) :: this
  logical :: has_rew
  has_rew = this%has_reweighting
  return
end function

subroutine new_wilson_quark_phmc(this,id)
  implicit none
  type(hmc_poly_quark_wilson), intent(inout) :: this
  integer, intent(in) :: id
  character(len=CHARLEN) :: str
  NULLIFY(this%y)

  this%myid = id
  call new(this%phys_param,id)
  call new(this%update_q_inv)
  call new(this%gmp_inv)

  write(str,'(I2.2)')id
  this%solver_log_fname = TRIM(this%solver_log_fname)//TRIM(str)
  this%gmp_log_fname    = TRIM(this%gmp_log_fname)//TRIM(str)

  call new(this%solver_log,this%solver_log_fname)
  call new(this%gmp_status,this%gmp_log_fname,this%gmp_test_name,this%gmp_weight_name)

  return
end subroutine

subroutine delete_wilson_quark_phmc(this)
  implicit none
  type(hmc_poly_quark_wilson), intent(inout) :: this
  deallocate(this%y)
  call delete(this%phys_param)
  call delete(this%solver_log)
  call delete(this%gmp_status)
  NULLIFY(this%y)
  return
end subroutine

subroutine read_wilson_quark_phmc(this,iout)
!
! Read PHMC quark algorithm parameters and Initialize algorithm parameters(ex. polynomial coeffs)
!
  use comlib
  implicit none
  type(hmc_poly_quark_wilson), intent(inout) :: this
  integer, intent(in) :: iout
  real(DP) :: rtmp0
  integer :: imb

  allocate(this%y)

  call new(this%update_q_inv)
  call new(this%gmp_inv)

  call read(this%phys_param,iout)
  if (nodeid==0) then
  read(iout,*)
  read(iout,*)this%update_q_inv%tol
  read(iout,*)this%gmp_inv%tol
  read(iout,*)this%depth_hopping
  read(iout,*)this%depth_clvtrlog
  read(iout,*)this%ptype
  read(iout,*)this%npoly
  read(iout,*)this%switch_metropolis_test
  endif
#ifndef _singlePU
  call comlib_bcast(this%update_q_inv%tol,0)
  call comlib_bcast(this%gmp_inv%tol,0)
  call comlib_bcast(this%depth_hopping,0)
  call comlib_bcast(this%depth_clvtrlog,0)
  call comlib_bcast(this%ptype,0)
  call comlib_bcast(this%npoly,0)
  call comlib_bcast(this%switch_metropolis_test,0)
#endif

  allocate(this%pcoef(0:this%npoly))
  allocate(this%zcoef(0:this%npoly))

  this%nflavor = get_nflavor(this%phys_param)

  select case (this%ptype)
  case (0) ! Chebyshev (c_i=(-1)^i)
    this%pcoef(0)=1.0_DP
    do imb=1,this%npoly
      this%pcoef(imb)=-this%pcoef(imb-1)
    enddo
    select case (this%nflavor)
    case (1) ! for 1 flavor
      call sqrtpoly_chev(this%npoly,this%pcoef,this%zcoef)
    case (2) ! for 2 flavor
      do imb=0,this%npoly
        rtmp0=this%pcoef(imb)
        this%zcoef(imb)=cmplx(rtmp0,0.0_DP,kind=DP)
      enddo
    case default
      call error_stop(" Error in quark param file (nflavor)")
    end select

  case (1) ! Adopted
    if (nodeid==0) then
    do imb=0,this%npoly
      read(iout,*)this%pcoef(imb)
    enddo
    endif
#ifndef _singlePU
    do imb=0,this%npoly
      call comlib_bcast(this%pcoef(imb),0)
    enddo
#endif
    select case (this%nflavor)
    case (1) ! for 1 flavor
      call sqrtpoly(this%npoly,this%pcoef,this%zcoef)
    case (2) ! for 2 flavor
      do imb=0,this%npoly
        rtmp0=this%pcoef(imb)
        this%zcoef(imb)=cmplx(rtmp0,0.0_DP,kind=DP)
      enddo
    case default
      call error_stop(" Error in quark param file (nflavor)")
    end select
  case default
    call error_stop(" Error in quark param file (ptype)")
  end select

  this%nsqrt=2

  return
end subroutine

subroutine save_config_hmc_poly_quark_wilson(this,iout)
!
! Save Quark parameter on measurement configuration
!
  implicit none
  type(hmc_poly_quark_wilson), intent(in) :: this
  integer, intent(in) :: iout
  call save_config(this%phys_param,iout)
  return
end subroutine

subroutine print_wilson_quark_phmc(this)
!
! print out quark PHMC algorithm parameters on display
!
  implicit none
  type(hmc_poly_quark_wilson), intent(inout) :: this
  integer :: imb

  if (nodeid==0) then

  write(*,'("**** Quark # ",I3," PHMC")')this%myid

  write(*,'("***** Physical parameters *****")')
  call print(this%phys_param)
  write(*,'("***** PHMC Algorithm parameters *****")')
  write(*,'("    Heat-Bath Stopping Condition:",E16.8," (pseudo-fermion)")')this%update_q_inv%tol
  write(*,'("  MD force depth PSfermion part :",I3)')this%depth_hopping
#ifdef _CLOVER
  write(*,'("  MD force depth CLV trlog part :",I3)')this%depth_clvtrlog
#endif

  select case (this%switch_metropolis_test)
  case (0)
    write(*,'("              Global Metropolis : No ")')
  case (1)
    write(*,'("              Global Metropolis : Yes")')
    write(*,'("  Metropolis Stopping Condition :",E16.8)') this%gmp_inv%tol
  case default
    call error_stop("Metropolis test switch error")
  end select

  write(*,'(14X," Polynomial Order :",I4)') this%npoly

  select case (this%ptype)
  case (0)
    write(*,'(16X,"Polynomial type : Chebyshev")')
  case (1)
    write(*,'(16X,"Polynomial type : Adopted")')
  case default
    call error_stop("Polynomial type error")
  end select

  write(*,'("        Polynomial Coefficients :")')
  write(*,'(4X,7X,"c_i (real)")')
  write(*,'(80("-"))')
  do imb=0,this%npoly
    write(*,'(I4,E24.16)') imb,this%pcoef(imb)
  enddo
  write(*,'("  Force Polynomial Coefficients :")')
  write(*,'(4X,7X,"d_i (complex)")')
  write(*,'(80("-"))')
  do imb=0,this%npoly/(3-this%nflavor)
    write(*,'(I4,2E24.16)') imb,this%zcoef(imb)
  enddo
  write(*,'(80("-"))')

  endif

  return
end subroutine

subroutine print_stat_wilson_quark_phmc(this)
!
! print out quark PHMC algorithm job statistics on display
!
  implicit none
  type(hmc_poly_quark_wilson), intent(inout) :: this
  character(CHARLEN) :: name
  real(DP) :: hb_ave,gmp_ave

  hb_ave = get_average(this%update_q_inv)
  if (nodeid==0) then
  write(*,'("**** Quark # ",I3," PHMC")')this%myid

  call print(this%phys_param)

  write(*,'(" PHMC algorithm ")')
  write(*,'(" Heat-Bath for Pseudofermion BiCGStab iter")')
  write(*,'("                       (averaged) :",F12.4)') hb_ave
  endif

  if (SW_ON == this%switch_metropolis_test) then

    call print_statistics(this%gmp_status)

    if (this%nflavor == 2) then
      gmp_ave = get_average(this%gmp_inv)
      if (nodeid==0) then
      write(*,'(" Global Metropolis BiCGStab iter")')
      write(*,'("                       (averaged) :",F12.4)')gmp_ave
      endif
    endif
  endif

  return
end subroutine


subroutine force_wilson_quark_phmc(this,BB,idepth,u)
!
! Calc MD force from hoping part of the even/odd-site preconditioned Polynomial HMC action.
!
!   Action : Sphmc
!                            v
!     Sphm : PHMC action, |P[Doo]yo|^2  (odd site only)
!                  v
!            where Doo = 1oo - kappa^2 Fo^-1 Moe Fe^-1 Meo
! 
 
  implicit none
  type(hmc_poly_quark_wilson), intent(inout) :: this    ! quark action/algorithm parameters
  integer,                     intent(in)    :: idepth  ! MD depth for recuresive MD integrator
  type(vfield_gluon_wg),       intent(in)    :: u       ! gauge field
  type(vfield_gluon_wg),       intent(inout) :: BB      ! pre-force contribution (\dot{u})

  real(DP) :: kappa2,fcoef
  type(field_quark_wg),    allocatable :: fy(:),fx,fz
  type(field_quark_eo_wg), allocatable :: xo,wo
#ifdef _CLOVER
  real(DP) :: dummy
  real(DP) :: csw
  complex(DP) :: zickd8
#endif
  integer :: imb,nmb
  complex(DP) :: ccoef
  complex(DP), allocatable :: zcoef(:)
  integer :: nf
  real(DP) :: kappa


#ifdef _CLOVER
  if (idepth == this%depth_hopping .or. &
 &    idepth == this%depth_clvtrlog) then
  !=======================================================
  ! comp. inverce clover term matrix
  ! fclinv = [1 - csw kappa/2 sigma_{mu,nu}F_{mu,nu}]^-1
  !=======================================================
    call make_inverse_clover_term(this%phys_param,0,dummy)
  endif
#endif

  if (idepth == this%depth_hopping) then
#ifdef _MD_DEPTH_
    if (nodeid==0) write(*,'("depth=",I3," : force quark poly hopping")')idepth
#endif

  nf    = get_nflavor(this%phys_param)
  kappa =   get_kappa(this%phys_param)

  nmb=this%npoly/(3-nf)

  allocate(zcoef(0:nmb))
  allocate(fy(0:nmb),fx,fz,xo,wo)
  do imb=0,nmb
    call new(fy(imb)%eo(0),0)
    call new(fy(imb)%eo(1),1)
  enddo

  kappa2 = kappa**2
  fcoef  = -kappa2
  do imb=1,nmb
    zcoef(imb)=-(this%zcoef(imb)/this%zcoef(imb-1))*kappa2
  enddo
#ifdef _CLOVER
    csw  =     get_csw(this%phys_param)
  zickd8 = -(zi*csw*kappa**3)/8.0_DP
#endif

!======================================================================
! Prepare gauge force contributions from polynomial pseudo ferimon
!======================================================================
  !===========================
  !  fyo(nmb) <= yo
  !===========================
  call assign(fy(nmb)%eo(1),this%y%eo(1))

  do imb=nmb,1,-1
    !*************************************************************
    !*                              v
    !*  fyo^(j-1) = fyo^(N) + zcoef Moo fyo^(j), for j=N,N-1,...,1
    !*              _
    !*  fye^(j-1) = Meo fyo^(j-1)
    !*
    !*************************************************************
    call assign_mult_hop(this%phys_param,fy(imb  )%eo(0),fy(imb)%eo(1),u)
    call assign_mult_hop(this%phys_param,fy(imb-1)%eo(1),fy(imb)%eo(0),u)
    !===========================================
    !  fyo(j-1) <= fyo(j-1)*zeof(j) + fyo(N)
    !===========================================
    call accum_mult_add(fy(imb-1)%eo(1),zcoef(imb),fy(nmb)%eo(1))
  enddo ! end of do imb

!***************************************
!* xo := F^-1oo gamma_5 fyo^(0)
!***************************************
#ifdef _CLOVER
  call assign_mult_gamma5(xo,fy(0)%eo(1))
  call accum_mult_clover_term(xo,this%phys_param)
#else
  call assign_mult_gamma5(xo,fy(0)%eo(1))
#endif

!************************************************************************
!* Force loop
!************************************************************************
  do imb=1,nmb ! polynomial loop
!*************************************************************
!*
!* xo =  (c_1/c_0)^* xo    (for imb=1)
!*                           v
!* xo =  (c_imb/c_{imb-1})^* Moo xo    (for imb=2,3,...,N)
!*
!* fxo = gamma_5 xo
!*               _
!* fxe = gamma_5 Meo xo
!*
!*************************************************************
    select case (imb)
    case (1)
      ccoef=conjg(this%zcoef(1)/this%zcoef(0))
    case (2:)
      ccoef=conjg(zcoef(imb))
    end select
    !====================
    !  xo <= xo * ccoef
    !====================
    call accum_mult(xo,ccoef)

    call assign_mult_hop(this%phys_param,this%y%eo(0),xo,u)

    call assign_mult_gamma5(fx%eo(1),xo)
    call assign_mult_gamma5(fx%eo(0),this%y%eo(0))

    call copy_boundary(fx%eo(1))
    call copy_boundary(fx%eo(0))

    if (imb < nmb) then
      call assign_mult_hop(this%phys_param,xo,this%y%eo(0),u)
    endif

!======================================
! force from hopping matrix
!======================================
    call force_hmc_hopping(BB,fcoef,fx,fy(imb))

#ifdef _CLOVER
    call assign(fz%eo(0),fy(imb)%eo(0))
    call assign_mult_hop(this%phys_param,fz%eo(1),fz%eo(0),u)
!======================================
! force from clover term in hopping
!======================================
    call force_hmc_clover(this%XX,zickd8,fx,fz)
#endif

  enddo ! end of do imb

  deallocate(zcoef)
  deallocate(fx,fy,fz,xo,wo)
  endif

#ifdef _CLOVER
  if (idepth == this%depth_clvtrlog) then
#ifdef _MD_DEPTH_
    if (nodeid==0) write(*,'("depth=",I3," : force quark poly trlog")')idepth
#endif
!======================================
! calc force from -Nf*log(det[F])
!======================================
    call force_clover_trlog(this%phys_param,this%XX)
  endif
  call delete_inverse_clover_term(this%phys_param)

!======================================
! unlink clover term contribution
! XX is used for interfacing.
!======================================
  this%XX => NULL()
#endif 

  return
end subroutine

subroutine hamil_wilson_quark_phmc(this,ifi,u,action)
!
! Calculate Hamiltonian for PHMC action
!
! Action :
!             v 
!  Sphmc = |P[Doo] yo|^2 - 2 log[det[F]] for Nf=2, (yo:lives on odd site only)
!
!    or
!             v
!        = |T[Doo] yo|^2 - 1 log[det[F]] for Nf=1, (yo:lives on odd site only)
!
!   where 
! 
!   P[x] = (1/c_0)Sum[c_i x^i,i=0,npoly],
! 
!   and
! 
!   T[x] = (1/d_0)Sum[d_i x^i,i=0,npoly/2].
! 
  use comlib
  implicit none
  type(hmc_poly_quark_wilson), &
                         intent(inout) :: this   ! quark action/algorithm paramters
  integer,               intent(in)    :: ifi    ! target switch, 
                                                 ! * 0 for initial with generating PF field,
                                                 ! * 1 for final action,
                                                 ! * 2 for revcheck
  type(vfield_gluon_wg), intent(in)    :: u      ! gauge field
  real(DP),              intent(out)   :: action ! action value

  type(field_quark_eo_wg), allocatable :: wo,ro
  integer :: imb,nmb,iiter
  integer :: nf
  real(DP) :: Spf,Sclv
  real(DP) :: err,  kappa
  complex(DP), allocatable :: zcoef(:)
  character(CHARLEN) :: str
  type(hmc_status) :: status

  nf    = get_nflavor(this%phys_param)
  kappa =   get_kappa(this%phys_param)

  Spf =0.0_DP
  Sclv=0.0_DP
#ifdef _CLOVER
  !===========================================
  !  calc. F^-1 and
  !  actaion value of clover term trace log
  !===========================================
  call hamil_clover_trlog(this%phys_param,Sclv)
  if (0 == this%depth_clvtrlog) Sclv = 0.0_DP
#endif


  select case(ifi)
  case(0) ! INITIAL

    write(str,'("# HAMIL SOLVER")')
    call print(this%solver_log,TRIM(str))

    nmb = this%npoly/(3-nf)
    allocate(zcoef(0:nmb))
    allocate(wo,ro)
    call new(wo,1)
    call new(ro,1)
    call set_gaussian_noise(wo)
    !=============================
    ! calc. pseudo fermion action
    !=============================
    Spf=abs2(wo)

    !=============================
    !            v
    !  Solve : P[Doo] yo = wo
    ! 
    ! Using preconditioning. Solve
    !            v      
    !          W[Doo] ro = wo
    !            v
    !   =>  yo = Doo ro
    !=============================
    err = this%update_q_inv%tol
    call assign_inv_mult_poly_prec(this,err,iiter,ro,wo,u)
    call inc(this%update_q_inv,iiter)

    call assign_mult_eoprec_wd(this%phys_param,this%y%eo(1),ro,u)

    !=================================================
    ! 
    !  For Nf=1,
    ! 
    !   yo <= T[D]^* yo
    !                                            v                            v
    !      = (1/d^*_0) [ d^*_0 + d^*_1 (-kappa^2 Moo) + ... + d^*_n (-kappa^2 Moo)^n ] yo
    !                                  v                             v
    !      = [ 1 - kappa^2 (d_1/d_0)^* Moo [ 1 - kappa^2 (d_2/d_1)^* Moo [ ....
    !                   v                                 v
    !             ..... Moo [ 1 - kappa^2 (d_n/d_{n-1})^* Moo ]...] yo
    !=================================================
    if (1 == nf) then

      do imb=1,this%npoly/2
        zcoef(imb)=-conjg((this%zcoef(imb)/this%zcoef(imb-1)))*(kappa**2)
      enddo

      call assign(wo,this%y%eo(1))

      do imb=this%npoly/2,1,-1
        call assign_mult_hop(this%phys_param,this%y%eo(0),          wo,u)
        call assign_mult_hop(this%phys_param,          wo,this%y%eo(0),u)
        call accum_mult_add(wo,zcoef(imb),this%y%eo(1))
      enddo ! end of do imb

      call assign(this%y%eo(1),wo)
    endif

    deallocate(zcoef,wo,ro)

  case(1:2) ! FINAL : REVCHECK 
    nmb=this%npoly/(3-nf)
    allocate(zcoef(0:nmb))
    allocate(wo)
    call new(wo,1)

    !====================
    !        v
    ! wo = P[Doo] yo
    !
    ! or
    !        v
    ! wo = T[Doo] yo
    !====================
    do imb=1,nmb
      zcoef(imb)=-(this%zcoef(imb)/this%zcoef(imb-1))*(kappa**2)
    enddo

    call assign(wo,this%y%eo(1))

    do imb=nmb,1,-1
      call assign_mult_hop(this%phys_param,this%y%eo(0),          wo,u)
      call assign_mult_hop(this%phys_param,          wo,this%y%eo(0),u)
      call accum_mult_add(wo,zcoef(imb),this%y%eo(1))
    enddo

    Spf=abs2(wo)
    deallocate(zcoef,wo)
  end select

  if (0 == this%depth_hopping) Spf = 0.0_DP

#ifdef _CLOVER
  call delete_inverse_clover_term(this%phys_param)
#endif

  action=Spf+Sclv

  write(str,'("@",I8,I2," Quark: ID:",I3," SQPF:",E24.16," SCLV:",E24.16)') &
 &  get_trajectory_number(status),ifi,this%myid,Spf,Sclv
  call print_log_action(status,TRIM(str))

  return
end subroutine

subroutine reweighting_wilson_quark_phmc(this,u)
!
! Nothing is done
!
  implicit none
  type(hmc_poly_quark_wilson), intent(inout) :: this ! quark action/algorihtm parameters
  type(vfield_gluon_wg),       intent(in)    :: u
  return
end subroutine

subroutine metropolis_wilson_quark_phmc(this,u1,u0,ihit)
!
! Do global Metropolis test for PHMC action (correction factor)
!
  implicit none
  type(hmc_poly_quark_wilson), intent(inout) :: this ! quark action/algorihtm parameters
  type(vfield_gluon_wg),       intent(in)    :: u1   ! MD final gauge field
  type(vfield_gluon_wg),       intent(in)    :: u0   ! MD inital gauge field
  integer,                     intent(inout) :: ihit ! 1 for accept, 0 for reject
  integer :: nf

  nf = get_nflavor(this%phys_param)

  if (this%switch_metropolis_test == SW_OFF) return
  if (ihit == 0) return
  select case (nf)
  case (1)
    call metropolis_nf1_phmc(this,u1,u0,ihit)
  case (2)
    call metropolis_nf2_phmc(this,u1,u0,ihit)
  case default
    call error_stop(" # of flavor exceeds 2.")
  end select

  return
end subroutine

subroutine metropolis_nf1_phmc(this,u1,u0,ihit)
!
! noisy Metropolis test for odd number flavor quarks with 
!                            v
!    PHMC correction factor (W[Doo] odd site)
!
  use comlib
  use random_class
  implicit none
  type(hmc_poly_quark_wilson), intent(inout) :: this
  type(vfield_gluon_wg),       intent(in)    :: u1,u0
  integer ,                    intent(inout) :: ihit

  type(field_quark_eo_wg), allocatable :: wo(:)
  type(tfield_gluon_wg), allocatable :: wl34,wl98
  real(DP) :: rtmp0,rtmp1,dummy,err,ds
  logical :: flag
  integer :: itraj
  integer :: iiter,niter,i
  character(CHARLEN) :: str
  type(hmc_status) :: status

  if (ihit == 0) return

  itraj = get_trajectory_number(status)

  allocate(wo(6))
  do i=1,6
  call new(wo(i),1)
  enddo

  niter=100
  err = this%gmp_inv%tol

  if (nodeid == 0) then
    write(str,'(" Taylor Expansion Approximation for Square root of correction Matrix")')
    call print_log(this%gmp_status,str)
    write(str,'("       Max iteration :",I3)') niter
    call print_log(this%gmp_status,str)
    write(str,'("  Stopping condition :",E13.6)')err
    call print_log(this%gmp_status,str)
  endif

!***************************************
!*  w1 = gaussian noise (eta)
!***************************************
  call set_gaussian_noise(wo(1))

  iiter=1

!***************************************
!* 
!*  w2 <= A w1
!*
!* where
!*
!*   A = (1+delta)^(1/2)
!*
!* with
!*             v 
!*   delta = W[D] - 1,
!*          v      v  v
!* where  W[D] = P[D] D.
!*
!***************************************
#ifdef _CLOVER
  allocate(wl34,wl98)
  call make_two_links(u0,wl34,wl98)
  call make_clover_leaf(u0,wl34,wl98)
  deallocate(wl34,wl98)
  call make_inverse_clover_term(this%phys_param,0,dummy)
  call delete_clover_leaf()
#endif
  call assign_mult_poly_prec(this,wo(5),wo(1),u0)
  call accum_mult(wo(5),this%pcoef(0))

100   continue
!***************************************
  call assign_mult_sqrt_poly_prec_taylor(this,+1,wo(2),wo(1),u0)
  call assign_mult_sqrt_poly_prec_taylor(this,+1,wo(4),wo(2),u0)

!***************************************
!* calc residulal
!*     eta = w1o
!*   W eta = w5o
!*   A eta = w2o
!* A A eta = w4o
!*
!*     || W eta - A A eta ||
!* r = ---------------------
!*         || W eta ||
!***************************************
  !======================
  ! wo(6) = wo(5)-wo(4)
  !======================
  call assign_sub(wo(6),wo(5),wo(4))
  rtmp0=abs2(wo(6))
  rtmp1=abs2(wo(5))
  rtmp0=sqrt(rtmp0/rtmp1)
  if (nodeid == 0) then
    write(str,'("# for sqrt[W_corr]")')
    call print_log(this%gmp_status,str)
    write(str,'("#",2I5," ERR:",E24.16)')iiter,this%nsqrt,rtmp0
    call print_log(this%gmp_status,str)
  endif
  if (rtmp0 > err) then
    this%nsqrt=this%nsqrt+1
    iiter=iiter+1
    if (iiter >= niter) then
      if (nodeid == 0) then
        write(str,'("#",I5," ERR:",E24.16)')iiter,rtmp0
        call print_log(this%gmp_status,str)
        write(str,'("# Taylor expansion does not converge! : Stop")')
        call print_log(this%gmp_status,str)
      endif 
      call error_stop("sqrt W solver did not converge.")
    else
      goto 100
    endif
  endif

  this%nsqrt=this%nsqrt+1
  iiter=1
!***************************************
!*
!*  w3 = A' w2
!*
!* where
!*
!*   A' = (1+delta')^(-1/2)
!*
!* with
!*              v 
!*   delta' = W[D'] - 1
!*
!***************************************
#ifdef _CLOVER
  call delete_inverse_clover_term(this%phys_param)
  allocate(wl34,wl98)
  call make_two_links(u1,wl34,wl98)
  call make_clover_leaf(u1,wl34,wl98)
  deallocate(wl34,wl98)
  call make_inverse_clover_term(this%phys_param,0,dummy)
  call delete_clover_leaf()
#endif

200   continue
!***************************************
  call assign_mult_sqrt_poly_prec_taylor(this,-1,wo(3),wo(2),u1)
  call assign_mult_sqrt_poly_prec_taylor(this,-1,wo(4),wo(3),u1)
  call assign_mult_poly_prec(this,wo(5),wo(4),u1)
  call accum_mult(wo(5),this%pcoef(0))

!***************************************
!* calc residulal
!*             chi = w2o
!*        A^-1 chi = w3o
!*   A^-1 A^-1 chi = w4o
!* W A^-1 A^-1 chi = w5o
!*
!*     || W A^-1 A^-1 chi - chi ||
!* r = ----------------------------
!*             || chi ||
!***************************************
  call assign_sub(wo(6),wo(5),wo(2))
  rtmp0=abs2(wo(6))
  rtmp1=abs2(wo(2))
  rtmp0=sqrt(rtmp0/rtmp1)
  if (nodeid == 0) then
    write(str,'("# for sqrt[W_corr]^-1")')
    call print_log(this%gmp_status,str)
    write(str,'("#",2I5," ERR:",E24.16)')iiter,this%nsqrt,rtmp0
    call print_log(this%gmp_status,str)
  endif
  if (rtmp0 > err) then
    this%nsqrt=this%nsqrt+1
    iiter=iiter+1
    if (iiter >= niter) then
      if (nodeid == 0) then
        write(str,'("#",I5," ERR:",E24.16)')iiter,rtmp0
        call print_log(this%gmp_status,str)
        write(str,'("# Taylor expansion does not converge! : Stop")')
        call print_log(this%gmp_status,str)
      endif 
      call error_stop("sqrt W^-1 solver did not converge.")
    else
      goto 200
    endif
  endif

!***************************************
!*
!*   ds = |w3|^2 - |w1|^2
!*
!***************************************
  ds = abs2(wo(3))-abs2(wo(1))

!****************************************
!* Metropolis test
!*  P[u0->u] = min(1,exp(-ds))
!****************************************
  flag = is_accepted(this%gmp_status,itraj,ds, .TRUE. )

  if (flag) then
    ihit = 1
  else
    ihit = 0
  endif

  this%nsqrt=this%nsqrt-1

#ifdef _CLOVER
  call delete_inverse_clover_term(this%phys_param)
#endif
  deallocate(wo)

  return
end subroutine

subroutine metropolis_nf2_phmc(this,u1,u0,ihit)
!
! noisy Metropolis test for even number flavor quarks with 
!
!   PHMC correction factor (W[Doo] odd site)
!
  use comlib
  use random_class
  implicit none
  type(hmc_poly_quark_wilson), intent(inout) :: this
  type(vfield_gluon_wg),       intent(in)    :: u1,u0
  integer ,                    intent(inout) :: ihit

  type(field_quark_eo_wg), allocatable :: w1o,w2o,ro
  type(tfield_gluon_wg),   allocatable :: wl34,wl98
  real(DP) :: ds,err,dummy
  character(len=CHARLEN) :: str
  logical :: flag
  integer :: iiter,itraj
  type(hmc_status) :: status

  if (ihit == 0) return

  itraj = get_trajectory_number(status)

  write(str,'("# METROPOLIS NF=2 SOLVER")')
  call print(this%solver_log,TRIM(str))

  allocate(w1o,w2o,ro)
  call new(w1o,1)
  call new(w2o,1)
  call new( ro,1)

!***************************************
!*  w1 = gaussian noise
!***************************************
  call set_gaussian_noise(w1o)

!***************************************
!*         v
!*   r = W[D] w1
!*          v      v v
!* where  W[D] = P[D]D
!*
!***************************************
#ifdef _CLOVER
  allocate(wl34,wl98)
  call make_two_links(u0,wl34,wl98)
  call make_clover_leaf(u0,wl34,wl98)
  deallocate(wl34,wl98)
  call make_inverse_clover_term(this%phys_param,0,dummy)
  call delete_clover_leaf()
#endif
  call assign_mult_poly_prec(this,ro,w1o,u0)

!***************************************
!*  Solve
!*      v
!*    W[D'] w2 = r
!*
!*              v          v
!* then  w2 = W[D']^{-1} W[D] w1
!*
!***************************************
  err = this%gmp_inv%tol
#ifdef _CLOVER
  call delete_inverse_clover_term(this%phys_param)
  allocate(wl34,wl98)
  call make_two_links(u1,wl34,wl98)
  call make_clover_leaf(u1,wl34,wl98)
  deallocate(wl34,wl98)
  call make_inverse_clover_term(this%phys_param,0,dummy)
  call delete_clover_leaf()
#endif
  call assign_inv_mult_poly_prec(this,err,iiter,w2o,ro,u1)
  call inc(this%gmp_inv,iiter)

!***************************************
!*
!*   ds = |w2|^2 - |w1|^2
!*
!***************************************
  ds = abs2(w2o)-abs2(w1o)

!****************************************
!* Metropolis test
!*  P[u0->u] = min(1,exp(-ds))
!****************************************
  ihit=0
  flag = is_accepted(this%gmp_status,itraj,ds, .TRUE. )
  if (flag) then
    ihit = 1
  else
    ihit = 0
  endif

  deallocate(w1o,w2o,ro)
#ifdef _CLOVER
  call delete_inverse_clover_term(this%phys_param)
#endif

  return
end subroutine


subroutine assign_mult_poly_prec(this,yde,ye,u)
!                                              v
! Multiply preconditioned correction matrix W[Dee] :
!
!          v 
!  yde = W[Dee] ye,
!
!  v                                           v        v    v
!  Dee = 1ee - kappa^2 Fe^-1 Meo Fo^-1 Moe,  W[Dee] = P[Dee] Dee
! 
  implicit none
  type(hmc_poly_quark_wilson), intent(in)    :: this  ! quark action/algorihtm parameters
  type(field_quark_eo_wg),     intent(inout) :: yde   ! even/odd site fermion filed (output)
  type(field_quark_eo_wg),     intent(inout) ::  ye   ! even/odd site fermion filed (input)
  type(vfield_gluon_wg),       intent(in)    ::   u   ! gauge field

  type(field_quark_eo_wg), allocatable :: Myo,w1e,w2e
  real(DP) :: coef(0:this%npoly),kappa
  integer :: imb

  kappa = get_kappa(this%phys_param)
  do imb=1,this%npoly
    coef(imb)=-(this%pcoef(imb)/this%pcoef(imb-1))*(kappa)**2
  enddo

  allocate(w1e)
  call new(w1e,ye%ieo)

!**********************************************************
!*         v
!*   w1e = Dee ye
!**********************************************************
  call assign_mult_eoprec_wd(this%phys_param,w1e,ye,u)

!*******************************
!*   yd = w1
!*******************************
  yde=w1e

!**********************************************************
!*          v
!*  yde = P[D]ee w1e
!*
!*      = [ c0 + c1(D-1) + c2(D-1)^2 + ... + cn(D-1)^n ] w1e
!*
!*      = [ c0 + c1(-kappa^2 M^2) + c2(-kappa^2 M^2)^2 + 
!*                            ... + cn(-kappa^2 M^2)^n ] w1e
!*
!*      = c0 [ 1 - kappa^2 c1/c0 M^2 [ 1 - kappa^2 c2/c1 M^2 
!*                                ...[ 1 - kappa^2 cn/cn-1 M^2 ]...] w1e
!*
!**********************************************************
  allocate(Myo,w2e)

  do imb = this%npoly,1,-1
    !=============================
    ! Myo <= Mhop[U]oe yde
    ! w2e <= Mhop[U]eo Mdo
    !=============================
    call assign_mult_hop(this%phys_param,Myo,yde,u)
    call assign_mult_hop(this%phys_param,w2e,Myo,u)

    !=============================
    ! yde <= w2e * coef(imb) + w1e
    !=============================
    call assign_mult_add(yde,w2e,coef(imb),w1e)
  enddo ! end of do imb

  deallocate(Myo,w1e,w2e)

  return
end subroutine

subroutine assign_inv_mult_poly_prec(this,err,iiter,xe,be,u)
!
! Multiply inverse of polynomial preconditioned operator W[u] by solving :
!
!   W xe = be -> xe = W^-1 be,
!
! where
!  v                                           v        v    v
!  Dee = 1ee - kappa^2 Fe^-1 Meo Fo^-1 Moe,  W[Dee] = P[Dee] Dee
! 
  use timer_class
  use counter_class
  use solver_class
  implicit none
  type(hmc_poly_quark_wilson), intent(in)    :: this  ! quark action/algorithm parameters
  real(DP),                    intent(in)    :: err   ! solver stopping condition
  integer,                     intent(out)   :: iiter ! total number of iteration after convergence
  type(field_quark_eo_wg),     intent(inout) :: xe    ! even/odd site field output
  type(field_quark_eo_wg),     intent(in)    :: be    ! even/odd site field input (rhs)
  type(vfield_gluon_wg),       intent(in)    :: u     ! gauge field

  type(field_quark_eo_wg), allocatable :: we
  type(bicgstab_alg),      allocatable :: solver
  type(counter) :: imult
  integer    :: mult_count
  integer    :: iout,istat
  type(timer) :: solver_time
  real(DP) :: etime

  call new(solver_time)
  call tic(solver_time)

  iout = get_file_unit(this%solver_log)

  allocate(we,solver)
  call new(xe,be%ieo)
  call new(we,be%ieo)

  call new(solver,NSIZE=COL*SPIN*NTH*NZ*NY*NX,max_iter=100,guess=1,tol=err)
  imult = mult_iter

  call pack(be,solver%src_vec)
  solver%dst_vec(:) = solver%src_vec(:)

  do 
    call solve(solver)

    istat = get_status(solver)
    select case (istat)
    case (OP_NOP)
      cycle
    case (OP_DO_MATVEC)
      call unpack(solver%src_vec,xe)
      call assign_mult_poly_prec(this,we,xe,u)
      call   pack(we,solver%dst_vec)
    case (OP_PRINT_STATUS)
      if (nodeid == 0) then
        write(iout,'("@",I5," ERR_=",E24.16)') &
 &                          get_current_iteration(solver),&
 &                              get_residual_norm(solver)
      endif
    case (OP_CONVERGED)
      if (nodeid == 0) then
        mult_count = get_count(mult_iter)-get_count(imult)
        write(iout,'("@",I5," ERR_=",E24.16)') &
 &                          get_current_iteration(solver),&
 &                              get_residual_norm(solver)
        write(iout,'("#",A," iteration converged.")')TRIM(get_name(solver))
        write(iout,'("#",I5," ERR_=",E24.16," MULT_=",I10)',advance='no')   &
 &                          get_current_iteration(solver),&
 &                              get_residual_norm(solver),mult_count

      endif
      iiter = get_current_iteration(solver)
      exit      
    case (OP_MAXITER_REACHED)
      if (nodeid == 0) then
        mult_count = get_count(mult_iter)-get_count(imult)
        write(iout,'("@",I5," ERR_=",E24.16)') &
 &                          get_current_iteration(solver),&
 &                              get_residual_norm(solver)
        write(iout,'("#",A," iteration does not converge.")')TRIM(get_name(solver))
        write(iout,'("#",I5," ERR_=",E24.16," MULT_=",I10)')     &
 &                          get_current_iteration(solver),&
 &                              get_residual_norm(solver),mult_count
      endif
      call error_stop("W^-1 solver did not converge.")
    case default
      call error_stop("something is wrong in W^-1 solver.")
    end select

  enddo

  call unpack(solver%dst_vec,xe)
  call delete(solver)
  deallocate(solver,we)
  call toc(solver_time)
  etime = get_elapse(solver_time)
  if (nodeid == 0) then
    write(iout,'(" ETIME_=",E24.16)')etime
  endif

  return
end subroutine

subroutine assign_mult_sqrt_poly_prec_taylor(this,isgn,yde,ye,u)
!
! Calculate square root of correction matrix W using Taylor expansion
!  approximation:
!
!    yd = W^(+1/2) y
!   or                             v    v
!    yd = W^(-1/2) y, where  W = P[Dee] Dee
! 
  implicit none
  type(hmc_poly_quark_wilson), intent(inout) :: this  ! quark action/algorithm parameters
  integer,                     intent(in)    :: isgn  ! sign switch, +1 for ^(+1/2), -1 for ^(-1/2)
  type(field_quark_eo_wg),     intent(inout) :: yde   ! even/odd site fermion field(output)
  type(field_quark_eo_wg),     intent(in)    ::  ye   ! even/odd site fermion field(rhs)
  type(vfield_gluon_wg),       intent(in)    ::   u   ! gauge field

  type(field_quark_eo_wg), allocatable :: w1e
  integer :: isqrt
  real(DP) :: asqrt

  allocate(w1e)
  call new(w1e,ye%ieo)

  !===========
  ! yde <= ye
  !===========
  call assign(yde,ye)

!***************************************
!* 
!*  w3 <= A w3
!*
!* where
!*
!*   A = (1+delta)^(+1/2)
!*     = 1 + 1/2 delta - 1/8 delta^2 + 1/16 delta^3 + ...,
!*  or
!*   A = (1+delta)^(-1/2)
!*     = 1 - 1/2 delta + 3/8 delta^2 - 5/16 delta^3 + 
!* with
!*             v v       v
!*   delta = P[D]D exp[a M] - 1.
!*
!***************************************
  do isqrt=this%nsqrt,1,-1
    select case(isgn)
    case (+1)
      asqrt=real(3-2*isqrt,kind=DP)/real(2*isqrt,kind=DP)
    case (-1)
      asqrt=real(1-2*isqrt,kind=DP)/real(2*isqrt,kind=DP)
    case default
      call error_stop("isgn for sqrt of correction matrix must be 0 or 1.")
    end select
    call assign_mult_poly_prec(this,w1e,yde,u)

    call accum_mult(w1e,this%pcoef(0))      ! w1e <= w1e * pcoef
    call accum_sub(w1e,yde)                 ! w1e <= w1e - yde
    call assign_mult_add(yde,w1e,asqrt,ye)  ! yde <= w1e * asqrt + ye
  enddo

  deallocate(w1e)

  return
end subroutine

subroutine sqrtpoly(np,pcoef,zcoef)
!
! Calc. polynomial coefficients for the square root of a polynomial :
!
! P[x] = sum[c_i x^i,i=0,np]    ! original polynomial : P[x]
!
!      = c_np prod[(x - z_k),k=1,np]
!
!      = c_np prod[(x - z_k),k=1,np/2] prod[(x - z^*_k),k=1,np/2]
!
!      = T[x] T'[x]             ! obtained polynomial : T[x]
!
!  where T[x] = sum[  d_i x^i,i=0,np/2],
!       T'[x] = sum[d^*_i x^i,i=0,np/2].
!
! I use bit-reversal ordering to divide complex roots into two groups.
!
  implicit none
  integer,      intent(in)  :: np            ! degree of polynomial
  real(DP),     intent(in)  :: pcoef(0:np)   ! polynomial coefficinets for P[x] = [pcoef(i)*x**i,i=0,np]
  complex(DP),  intent(out) :: zcoef(0:np/2) ! polynomial coefficients for T[x] = [zcoef(i)*x**i,i=0,np/2]

  complex(DP) :: zpcoef(np+1)
  complex(DP) :: zroot(np*2),zrootsort(np*2)
  real(DP) :: argz(np)
  integer :: indx(np)
  integer :: npow
  integer :: isort(0:np*2),ipow(0:np)
  integer :: i,k
  real(DP) :: rtmp0
  complex(DP), parameter :: z0=(0.0_DP,0.0_DP)

!*********** solve polynomial roots
      do i=0,np
        rtmp0=pcoef(i)
        zpcoef(i+1)=cmplx(rtmp0,0.0_DP,kind=DP)
      enddo
      call zroots(zpcoef,np,zroot,.true.)

!*********** sort polynomial roots 
      do i=1,np
        argz(i)=atan2(aimag(zroot(i)),real(zroot(i)))
      enddo
      call indexx(np,argz,indx)
      do i=1,np
        zrootsort(i)=zroot(indx(i))
      enddo

!******** bit reversal
      npow=(floor(log(REAL(np-1,kind=DP))/log(2.0_DP))+1)
      do i=0,2**npow-1

#ifndef __INTEL_COMPILER
!=========================================================
! Intel compiler XE 2011 -O2 produces incorrect results
!=========================================================
        do k=0,npow-1
          if (BTEST(i,k)) then
            ipow(k)= 1
          else
            ipow(k)= 0
          endif
        enddo
        isort(i)=0
        do k=0,npow-1
          isort(i)=isort(i)+ipow(npow-1-k)*2**k
        enddo
#else
        isort(i) = 0
        do k=0,npow-1
          if (BTEST(i,npow-1-k)) then
            isort(i)=isort(i)+2**k
          endif
        enddo
#endif

      enddo


      k=0
      do i=1,2**(npow-1)
        k=k+1
        if (k <= (np/2)) then
          zroot(i)=zrootsort(k)
        else
          zroot(i)=z0
        endif
      enddo
      k=np+1
      do i=2**npow,2**(npow-1)+1,-1
        k=k-1
        if (k >= (np/2+1)) then
          zroot(i)=zrootsort(k)
        else
          zroot(i)=z0
        endif
      enddo

!*********** take half roots
      k=0
      do i=1,2**(npow-1)
        if (zroot(isort(i-1)+1) /= z0) then
          k=k+1
          zrootsort(k)=zroot(isort(i-1)+1)
        endif
      enddo

!*********** solve polynomial coef.
      do i=0,np/2
        zcoef(i)=z0
      enddo
      zcoef(np/2)=sqrt(pcoef(np))

      do k=1,np/2
        do i=np/2-k,np/2-1
          zcoef(i)=zcoef(i)-zrootsort(k)*zcoef(i+1)
        enddo
      enddo

  return
end subroutine

subroutine sqrtpoly_chev(np,pcoef,zcoef)
!
! Calc. polynomial coefficients for the square root of the Chebyshev polynomial
!
! P[x] = sum[c_i x^i,i=0,np]    ! original polynomial : P[x]
!
!      = c_np prod[(x - z_k),k=1,np]
!
!      = c_np prod[(x - z_k),k=1,np/2] prod[(x - z^*_k),k=1,np/2]
!
!      = T[x] T'[x]             ! obtained polynomial : T[x]
!
!  where T[x] = sum[  d_i x^i,i=0,np/2],
!       T'[x] = sum[d^*_i x^i,i=0,np/2].
!
! I use bit-reversal ordering to divide complex roots into two groups.
!
  implicit none
  integer,      intent(in)  :: np            ! degree of polynomial
  real(DP),     intent(in)  :: pcoef(0:np)   ! polynomial coefficinets for P[x] = [pcoef(i)*x**i,i=0,np]
  complex(DP),  intent(out) :: zcoef(0:np/2) ! polynomial coefficients for T[x] = [zcoef(i)*x**i,i=0,np/2]

  complex(DP) :: zroot(np*2),zrootsort(np*2)
  real(DP) :: argz(np)
  integer :: indx(np)
  integer :: isort(0:np*2),ipow(0:np)
  integer :: npow
  integer :: i,k
  real(DP) :: rtmp0

!*********** Chebyshev polynomial roots
      do i=1,np
        rtmp0 = (8.0_DP*ATAN(1.0_DP)*i)/REAL(np+1,kind=DP)
        zroot(i) = cmplx(-COS(rtmp0),SIN(rtmp0),kind=DP)
      enddo

!*********** sort polynomial roots 
      do i=1,np
        argz(i) = ATAN2(AIMAG(zroot(i)),REAL(zroot(i),kind=DP))
      enddo

      call indexx(np,argz,indx)
      do i=1,np
        zrootsort(i) = zroot(indx(i))
      enddo

!******** bit reversal
      npow = (FLOOR(LOG(REAL(np-1,kind=DP))/LOG(2.0_DP))+1)
      do i=0,2**npow-1
#ifndef __INTEL_COMPILER
!=========================================================
! Intel compiler XE 2011 -O2 produces incorrect results
!=========================================================
        do k=0,npow-1
          if (BTEST(i,k)) then
            ipow(k) = 1
          else
            ipow(k) = 0
          endif
        enddo
        isort(i) = 0
        do k=0,npow-1
          isort(i)=isort(i)+ipow(npow-1-k)*2**k
        enddo
#else
        isort(i) = 0
        do k=0,npow-1
          if (BTEST(i,npow-1-k)) then
            isort(i)=isort(i)+2**k
          endif
        enddo
#endif
      enddo

      k=0
      do i=1,2**(npow-1)
        k=k+1
        if (k <= (np/2)) then
          zroot(i) = zrootsort(k)
        else
          zroot(i) = Z0
        endif
      enddo
      k=np+1
      do i=2**npow,2**(npow-1)+1,-1
        k=k-1
        if (k >= (np/2+1)) then
          zroot(i) = zrootsort(k)
        else
          zroot(i) = Z0
        endif
      enddo

!*********** take half roots
      k=0
      do i=1,2**(npow-1)
        if (zroot(isort(i-1)+1) /= Z0) then
          k=k+1
          zrootsort(k) = zroot(isort(i-1)+1)
        endif
      enddo

!*********** solve polynomial coef.
      do i=0,np/2
        zcoef(i) = Z0
      enddo
      zcoef(np/2) = SQRT(pcoef(np))

      do k=1,np/2
        do i=np/2-k,np/2-1
          zcoef(i) = zcoef(i) - zrootsort(k)*zcoef(i+1)
        enddo
      enddo

  return
end subroutine


!====================================================

SUBROUTINE zroots(a,m,roots,polish)
!
! Find polynomial roots. (to obtain single flavor PHMC coeffcients)
!
!  Modified from Neumerical recipes.
!
  IMPLICIT NONE
  INTEGER, INTENT(IN) :: m
  COMPLEX(KIND=DP), INTENT(IN) :: a(m+1)
  COMPLEX(KIND=DP), INTENT(OUT) :: roots(m)
  LOGICAL, INTENT(IN) :: polish
  REAL(KIND=DP) :: EPS
!CU    USES laguer
  INTEGER :: i,j,jj,its
  COMPLEX(KIND=DP) :: ad(m+1),x,b,c

  EPS=EPSILON(EPS)
  do j=1,m+1
    ad(j)=a(j)
  end do
  do j=m,1,-1
    x=(0.0_DP,0.0_DP)
    call laguer(ad,j,x,its)
    if (abs(aimag(x)) <= 2.0_DP*EPS**2*abs(real(x))) x=cmplx(real(x),0.0_DP,kind=DP)
    roots(j)=x
    b=ad(j+1)
    do jj=j,1,-1
      c=ad(jj)
      ad(jj)=b
      b=x*b+c
    end do
  end do
  if (polish) then
    do j=1,m
      call laguer(a,m,roots(j),its)
    end do
  endif
  do j=2,m
    x=roots(j)
    do i=j-1,1,-1
      if (real(roots(i)) <= real(x)) goto 10
      roots(i+1)=roots(i)
    end do
    i=0
10  roots(i+1)=x
  enddo

  return
END SUBROUTINE zroots


SUBROUTINE laguer(a,m,x,its)
!
! Laguerr method for finding polynomial roots
!
  IMPLICIT NONE
  INTEGER, INTENT(IN) :: m
  COMPLEX(KIND=DP), INTENT(IN) :: a(m+1)
  COMPLEX(KIND=DP), INTENT(INOUT) :: x
  INTEGER, INTENT(OUT) :: its
  INTEGER, PARAMETER :: MR=8,MT=500,MAXIT=MT*MR
  REAL(KIND=DP) :: EPSS
  INTEGER :: iter,j
  REAL(KIND=DP) :: abx,abp,abm,err,frac(MR)
  COMPLEX(KIND=DP) :: dx,x1,b,d,f,g,h,sq,gp,gm,g2
  SAVE frac
  DATA frac /.5_DP,.25_DP,.75_DP,.13_DP, &
 &           .38_DP,.62_DP,.88_DP,1._DP/

  EPSS=EPSILON(EPSS)*1.0_DP
  do iter=1,MAXIT
    its=iter
    b=a(m+1)
    err=abs(b)
    d=(0.0_DP,0.0_DP)
    f=(0.0_DP,0.0_DP)
    abx=abs(x)
    do j=m,1,-1
      f=x*f+d
      d=x*d+b
      b=x*b+a(j)
      err=abs(b)+abx*err
    end do
    err=EPSS*err
    if (abs(b) <= err) then
      return
    else
      g=d/b
      g2=g*g
      h=g2-2.*f/b
      sq=sqrt((m-1)*(m*h-g2))
      gp=g+sq
      gm=g-sq
      abp=abs(gp)
      abm=abs(gm)
      if (abp < abm) gp=gm
      if (max(abp,abm) > 0.0_DP) then
        dx=m/gp
      else
        dx=exp(cmplx(log(1.0_DP+abx),real(iter),kind=DP))
      end if
    end if
    x1=x-dx
    if (x == x1) return
    if (mod(iter,MT) /= 0) then
      x=x1
    else
      x=x-dx*frac(iter/MT)
    end if
  end do
  write(*,'("too many iterations in laguer")')

  return
END SUBROUTINE

subroutine indexx(n,arr,indx)
!
! Quick Sort (modified from Numerical Recipes in FORTRAN)
! to sort polynomial roots for single flavor PHMC.
!
  implicit none
  INTEGER,  intent(in)  :: n        ! number of data elements
  INTEGER,  intent(out) :: indx(n)  ! index array mapping original and sorted index 
  REAL(DP), intent(in)  :: arr(n)   ! target data array (to be indexed for sorting)

  INTEGER , PARAMETER  :: M=7,NSTACK=50
  INTEGER :: i,indxt,ir,itemp,j,jstack,k,l,istack(NSTACK)
  REAL(DP) :: a

      do j=1,n
        indx(j)=j
      enddo
      jstack=0
      l=1
      ir=n
1     if (ir-l < M) then
        do j=l+1,ir
          indxt=indx(j)
          a=arr(indxt)
          do i=j-1,1,-1
            if (arr(indx(i)) <= a) goto 2
            indx(i+1)=indx(i)
          enddo
          i=0
2         indx(i+1)=indxt
        enddo
        if (jstack == 0) return
        ir=istack(jstack)
        l=istack(jstack-1)
        jstack=jstack-2
      else
        k=(l+ir)/2
        itemp=indx(k)
        indx(k)=indx(l+1)
        indx(l+1)=itemp
        if (arr(indx(l+1)) > arr(indx(ir))) then
          itemp=indx(l+1)
          indx(l+1)=indx(ir)
          indx(ir)=itemp
        endif
        if (arr(indx(l)) > arr(indx(ir))) then
          itemp=indx(l)
          indx(l)=indx(ir)
          indx(ir)=itemp
        endif
        if (arr(indx(l+1)) > arr(indx(l))) then
          itemp=indx(l+1)
          indx(l+1)=indx(l)
          indx(l)=itemp
        endif
        i=l+1
        j=ir
        indxt=indx(l)
        a=arr(indxt)
3       continue
          i=i+1
        if (arr(indx(i)) < a) goto 3
4       continue
          j=j-1
        if (arr(indx(j)) > a) goto 4
        if (j < i) goto 5
        itemp=indx(i)
        indx(i)=indx(j)
        indx(j)=itemp
        goto 3
5       indx(l)=indx(j)
        indx(j)=indxt
        jstack=jstack+2
        if (jstack > NSTACK) then
          call error_stop("NSTACK too small in __FILE__.")
        endif
        if(ir-i+1 >= j-l)then
          istack(jstack)=ir
          istack(jstack-1)=i
          ir=j-1
        else
          istack(jstack)=j-1
          istack(jstack-1)=l
          l=i
        endif
      endif
      goto 1
end subroutine


end module
