#$Id: make.inc,v 1.4 2011/10/29 09:06:33 ishikawa Exp $
#########################################################################
### include file for Makefile
#########################################################################
DATE = '"$(shell env LC_ALL=C date +"%b %d %Y")"'
TIME = '"$(shell env LC_ALL=C date +"%H:%M:%S")"'
REVISION = -D_REVISION_='"$(shell cat ../.svnrevision)"'

#####################################################################
# Common section
#####################################################################
RM	= /bin/rm -f
MV	= /bin/mv


#####################################################################
# GPU Solver section
#####################################################################
GPU:=CUDA
#GPU:=NONE
ifeq ($(GPU),CUDA)
  USEGPU = -D_GPU_SOLVER_

  # Set CUDA
#  CUDADIR = $(HOME)/opt/cuda
  CUDADIR = /opt/cuda

  # Set CUDA SDK path
#  CUDASDKDIR = $(HOME)/opt/NVIDIA_GPU_Computing_SDK/C
  CUDASDKDIR = /opt/NVIDIA_GPU_Computing_SDK/C

  # Set CUDA Include
  CUDAINC = -I. -I$(CUDADIR)/include -I$(CUDASDKDIR)/common/inc

  # Set CUDA LIBS and GPU Solver LIBS
  CUDALIB = -L../GPUSolverClass -lgpusolver \
            -L../GPUBlockSolverClass -lgpublksolver \
            -L$(CUDADIR)/lib64 -L$(CUDASDKDIR)/lib -lcudart -lcutil_x86_64 -lm  -lstdc++

else
  USEGPU =
endif

#####################################################################
# Compiler Secion
#####################################################################
MACHINE := INTEL
#MACHINE := GCC

###################################
# Intel Fortran
###################################
ifeq ($(MACHINE),INTEL)

###################
# Archiver
###################
AR      = xiar
RANLIB  = ranlib

###################
# LAPACK95 PATH
###################
LAPACKINC = -I$(HOME)/include/lapack95_modules_ifort
LAPACKLIB = -L$(HOME)/lib -llapack95_ifort
#LAPACKINC = -I$(HOME)/include/lapack95_modules_intel
#LAPACKLIB = -L$(HOME)/lib -llapack95_intel

###################
# LAPACK PATH
###################
#MKLPATH=$(MKLROOT)/lib/em64t
MKLPATH=$(MKLROOT)/lib/intel64
LAPACKINC += -I$(MKLROOT)/include
LAPACKLIB += -L$(MKLPATH) \
          -Wl,--start-group $(MKLPATH)/libmkl_intel_lp64.a \
                            $(MKLPATH)/libmkl_intel_thread.a \
                            $(MKLPATH)/libmkl_core.a \
          -Wl,--end-group -liomp5 -lpthread

#################
# FFT LIBS
#################
FFTINC = -I$(HOME)/include
FFTLIB = $(HOME)/lib/libfftw3_intel.a

###################
# MPI compilers
###################
MPIHOME=/opt/openmpi-1.4.3
#MPIHOME=/usr/apps/openmpi/1.4.2/intel
FC      = $(MPIHOME)/bin/mpif90
#CC      = icc -O2 -align $(OPENMP)
CC      = icc -O2 -align
CPP     = cpp

OPENMP  = -openmp -openmp-report2
#OPENMP  = 

#DEBUG   = -p -g -traceback -D_DEBUG
#DEBUG   = -g -traceback -D_DEBUG -D_PERFMON_
#DEBUG   = -g -traceback -D_DEBUG

DEBUG   = -g -traceback -D_FLAT_

#DEBUG   = -g -traceback -check all -debug all  -D_FLAT_
#DEBUG   = -g -traceback -vec-report3 -opt-report
#DEBUG   = -check all -debug all -g -traceback -fpe-all=0 -fp-stack-check -D_DEBUG
#DEBUG   =  -fpe-all=0 -fp-stack-check -check all -debug all -g -traceback
#DEBUG   = -check all -g -traceback -fpe-all=0 -fp-stack-check -D_DEBUG
#DEBUG   = -g -traceback -fpe-all=0 -D_DEBUG
#DEBUG   = -p -g -traceback
#DEBUG   = -g -traceback -warn all
#FFLAGS  = -O2 -xSSE4.2 -stand f03 -i4 -u -static-intel -align all $(DEBUG) $(OPENMP) $(USEGPU)
#FFLAGS  = -O0 -xSSE4.2 -stand f03 -integer-size 32 -shared-intel -mcmodel=medium -align all $(DEBUG) $(OPENMP) $(USEGPU)

FFLAGS  = -O2 -xSSE4.2 -stand f03 -integer-size 32 -shared-intel -mcmodel=medium -align all \
          -finline -finline-limit=100000 -inline-factor=10000 -override-limits \
          $(DEBUG) $(OPENMP) $(USEQMX) $(USEGPU)

LDFLAGS = $(FFLAGS)

endif

###################################
# GFortran (GCC)
###################################
ifeq ($(MACHINE),GCC)

###################
# Archiver
###################
AR      = ar
RANLIB  = ranlib

########################
# LAPACK and LAPACK95
########################
LAPACKINC= -I$(HOME)/include/lapack95_modules_gcc
LAPACKLIB= -L$(HOME)/lib \
           -llapack95_gcc \
           -llapack_gcc \
           -lblas_gcc

#################
# FFT LIBS
#################
#FFTINC = -I$(HOME)/include
#FFTLIB = $(HOME)/lib/libfftw3_gcc.a

########################
# MPI Compiler
########################
MPIHOME=$(HOME)/opt/openmpi-1.4.5-gcc
FC      = $(MPIHOME)/bin/mpif90                  
CC      = $(MPIHOME)/bin/mpicc -O2 -falign-labels
#FC      = $(MPIHOME)/bin/mpif90                  
#CC      = $(MPIHOME)/bin/mpicc -O2 -falign-labels
#FC      = $(MPIHOME)/bin/mpif90                  
#CC      = $(MPIHOME)/bin/mpicc -O2 -falign-labels
CPP     = gcc
OPENMP  = -fopenmp
#OPENMP  = 
DEBUG   = -g -fbacktrace -D_FLAT_
#DEBUG   = -g -fbacktrace
#DEBUG   = -g -fbacktrace -fcheck=all -finit-local-zero
#DEBUG   = -g -fbacktrace -D_DEBUG
#DEBUG   = -g -fbacktrace -frange-check -fbounds-check -fvar-tracking -D_DEBUG
FFLAGS  = -O3 -msse4.2 -std=f2003 -finline-functions -falign-labels -fall-intrinsics -mcmodel=medium \
          $(DEBUG) $(OPENMP) $(USEQMX) $(USEGPU)
#FFLAGS  = -O3 -msse4.2  -finline-functions -falign-labels -fall-intrinsics \
#          $(DEBUG) $(OPENMP) $(USEQMX) $(USEGPU)
LDFLAGS = $(FFLAGS)

endif

#####################################################################
# QMX library section
#####################################################################
#COM:=QMX
COM:=NONE
ifeq ($(COM),QMX)
  USEQMX = -D_QMX_
  OMXPATH=/opt/open-mx-1.5.0
  QMXINC = -I../Qmxlib -I$(OMXPATH)/include
  QMXLIB = ../Qmxlib/qmx_lib.o ../Qmxlib/qmxf03.o -lrt $(OMXPATH)/lib/libmyriexpress.a -lstdc++
else
  USEQMX =
endif

#####################################################################
# Performance monitor PAPI section
#####################################################################
#PERFMON:=PAPI
PERFMON:=NONE
ifeq ($(PERFMON),PAPI)
  USEPAPI = -D_PERFMON
  PAPIDIR = /opt/papi-4.1.2.1
  PAPILIB = -L$(PAPIDIR)/lib -lpapi
  PAPIINC = -I$(PAPIDIR)/include
else
  USEPAPI =
endif


#####################################################################
# Summay section
#####################################################################

LIBS     = $(LAPACKLIB) $(FFTLIB) $(PAPILIB) $(QMXLIB) $(CUDALIB)
INCLUDES = $(LAPACKINC) $(FFTINC) $(PAPIINC) $(QMXINC) $(CUDAINC) \
           $(shell ls -1 -d ../*Class ../*Module | awk '{print "-I"$$1" "}')

OBJS    = $(SOURCES:.F90=.o)
LOGS    = $(SOURCES:.F90=.diag) $(SOURCES:.F90=.L) \
          $(SOURCES:.F90=.par)  $(SOURCES:.F90=.log) \
          $(SOURCES:.F90=.lst)

.PHONY: clean

.F90.o :

%.o : %.F90
	$(FC) $(FFLAGS) $(REVISION) $(MONITOR) $(INCLUDES) -c $< -o $@

%.s : %.F90
	$(FC) $(FFLAGS) $(REVISION) $(MONITOR) $(INCLUDES) -S $< -o $@

#########################################################################
